#include <iostream>
using namespace std;

#include "Macho.hpp"


// Simple microwave simulation
namespace Microwave {

    //////////////////////////////////////////////////////////////////////
    // State declarations

    // Machine's top state
    TOPSTATE(Top) {

        // Top state variables (visible to all substates)
        struct Box {
            Box() : myCookTime(0) {}
            void printTimer();
            void incrementTimer();
            void decrementTimer();
            void resetTimer();
            int getRemainingTime();
        private:
            int myCookTime;
        };

        STATE(Top)

        // Machine's event protocol
        virtual void open() {}
        virtual void close() {}
        virtual void minute() {}
        virtual void start() {}
        virtual void stop() {}
        virtual void tick() {}

    private:
        // special entry action
        void init();
    };

    // Microwave has been opened
    SUBSTATE(Disabled, Top) {
        STATE(Disabled)

        // Event handler
        void close();

    private:
        // Entry and exit actions of state
        void entry();
        void exit();
    };

    // Microwave is ready
    SUBSTATE(Operation, Top) {
        STATE(Operation)
        HISTORY()

        void open();
        void stop();

    private:
        void init();
    };

    // Microwave is idling
    SUBSTATE(Idle, Operation) {
        STATE(Idle)

        void minute();

    private:
        void entry();
    };

    // Microwave is being programmed
    SUBSTATE(Program, Operation) {
        STATE(Program)

        void minute();
        void start();
    };

    // Microwave is heating
    SUBSTATE(Cook, Operation) {
        STATE(Cook)

        void tick();
        void minute();

    private:
        void entry();
        void exit();
    };


    //////////////////////////////////////////////////////////////////////
    // Event handler and box implementations

    // Box of top state
    void Top::Box::printTimer() {
        cout << "  Timer set to " << myCookTime << " minutes" << endl;
    }
    void Top::Box::incrementTimer() {
        ++myCookTime;
    }
    void Top::Box::decrementTimer() {
        --myCookTime;
    }
    int Top::Box::getRemainingTime() {
        return myCookTime;
    }
    void Top::Box::resetTimer() {
        myCookTime = 0;
    }

    // Top state
    void Top::init() {
        setState<Operation>();
    }

    // State Operation
    void Operation::init() {
        setState<Idle>();
    }
    void Operation::open() {
        setState<Disabled>();
    }
    void Operation::stop() {
        setState<Idle>();
    }

    // State Disabled
    void Disabled::entry() {
        cout << "  Microwave open" << endl;
    }
    void Disabled::exit() {
        cout << "  Microwave closed" << endl;
    }
    void Disabled::close() {
        setState<Operation>();
    }

    // State Idle
    void Idle::entry() {
        TOP::box().resetTimer();
        cout << "  Microwave ready" << endl;
    }
    void Idle::minute() {
        setState<Program>();
        dispatch(Event(&TOP::minute));
    }

    // State Program
    void Program::minute() {
        TOP::box().incrementTimer();
        TOP::box().printTimer();
    }
    void Program::start() {
        setState<Cook>();
    }

    // State Cook
    void Cook::entry() {
        cout << "  Heating on" << endl;
    }
    void Cook::exit() {
        cout << "  Heating off" << endl;
    }
    void Cook::minute() {
        TOP::box().incrementTimer();
        TOP::box().printTimer();
    }
    void Cook::tick() {
        cout << "  Clock tick" << endl;
        TOP::box().decrementTimer();
        if (TOP::box().getRemainingTime() == 0) {
            cout << "  Finished" << endl;
            setState<Idle>();
        } else
            TOP::box().printTimer();
    }

} // namespace Microwave


int main() {
    Macho::Machine<Microwave::Top> m;

    cout << "Lets cook ourself a TV dinner:" << endl;

    m->minute();
    m->minute();
    m->minute();
    m->start();
    m->tick();
    m->open();
    cout << "Adding a little spice..." << endl;
    m->close();
    m->tick();
    m->tick();

    cout << "Now there is the remote...?" << endl;

    return 0;
}
