
// need to open up Machine class to make tests run
#define protected public
#define private public

// lets see everything that happens
#define __MACHO_TRACE__

#include "Macho.hpp"

#include <iostream>
using namespace std;

////////////////////////////////////////////////////////////////////////////////
// These tests will test for proper entry/exit, box and history handling
enum {
    STATE_TOP = 1 << 0,
    STATE_A = 1 << 1,
    STATE_AA = 1 << 2,
    STATE_AB = 1 << 3,
    STATE_AAA = 1 << 4,
    STATE_AAB = 1 << 5,
    STATE_ABA = 1 << 6,
    STATE_ABB = 1 << 7,
    STATE_B = 1 << 8,
    STATE_BA = 1 << 9,
    STATE_BB = 1 << 10,
    STATE_BAA = 1 << 11,
    STATE_BAB = 1 << 12,
    STATE_BBA = 1 << 13,
    STATE_BBB = 1 << 14,
    STATE_C = 1 << 15,
    STATE_CA = 1 << 16,
    STATE_CB = 1 << 17,
    STATE_CAA = 1 << 18,
    STATE_CAB = 1 << 19,
    STATE_CBA = 1 << 20,
    STATE_CBB = 1 << 21
};

namespace Transitions {

    TOPSTATE(Top) {
        struct Box {
            Box() : data(0) { clear(); }

            // Registering all exits and entries
            void entry(long state) { entries |= state; }
            void exit(long state) {
				exits |= state;
			}
            void clear() const { entries = 0; exits = 0; }

            mutable long entries;
            mutable long exits;
            mutable int data;
        };

        STATE(Top)

    private:
        void entry() { box().entry(STATE_TOP); }
        void exit() { box().exit(STATE_TOP); }
    };

    SUBSTATE(StateA, Top) {
        struct Box {
            Box() : data(0) {}
            int data;
        };

        STATE(StateA)

    private:
        // Testing proper box handling here
        void entry() { TOP::box().entry(STATE_A); box().data = 42; }
        void exit() { TOP::box().exit(STATE_A); assert(box().data == 42); }
    };

    SUBSTATE(StateAA, StateA) {
        STATE(StateAA)

    private:
        // Testing proper box handling here
        void entry() { TOP::box().entry(STATE_AA); assert(StateA::box().data == 42); StateA::box().data = 43; }
        void exit() { TOP::box().exit(STATE_AA); assert(StateA::box().data == 43); StateA::box().data = 42; }
    };

    SUBSTATE(StateAAA, StateAA) {
        STATE(StateAAA)

    private:
        void entry() { TOP::box().entry(STATE_AAA); }
        void exit() { TOP::box().exit(STATE_AAA); }
    };

    SUBSTATE(StateAAB, StateAA) {
        STATE(StateAAB)

    private:
        void entry() { TOP::box().entry(STATE_AAB); }
        void exit() { TOP::box().exit(STATE_AAB); }
    };

    SUBSTATE(StateAB, StateA) {
        STATE(StateAB)

    private:
        void entry() { TOP::box().entry(STATE_AB); }
        void exit() { TOP::box().exit(STATE_AB); }
    };

    SUBSTATE(StateABA, StateAB) {
        STATE(StateABA)

    private:
        void entry() { TOP::box().entry(STATE_ABA); }
        void exit() { TOP::box().exit(STATE_ABA); }
    };

    SUBSTATE(StateABB, StateAB) {
        STATE(StateABB)

    private:
        void entry() { TOP::box().entry(STATE_ABB); }
        void exit() { TOP::box().exit(STATE_ABB); }
    };

    SUBSTATE(StateB, Top) {
        STATE(StateB)
        HISTORY()

    private:
        void entry() { TOP::box().entry(STATE_B); }
        void exit() { TOP::box().exit(STATE_B); }
    };

    SUBSTATE(StateBA, StateB) {
        STATE(StateBA)

    private:
        void entry() { TOP::box().entry(STATE_BA); }
        void exit() { TOP::box().exit(STATE_BA); }
    };

    SUBSTATE(StateBAA, StateBA) {
        STATE(StateBAA)

    private:
        void entry() { TOP::box().entry(STATE_BAA); }
        void exit() { TOP::box().exit(STATE_BAA); }
    };

    SUBSTATE(StateBAB, StateBA) {
        STATE(StateBAB)

    private:
        void entry() { TOP::box().entry(STATE_BAB); }
        void exit() { TOP::box().exit(STATE_BAB); }
    };

    SUBSTATE(StateBB, StateB) {
        STATE(StateBB)

    private:
        void entry() { TOP::box().entry(STATE_BB); }
        void exit() { TOP::box().exit(STATE_BB); }
    };

    SUBSTATE(StateBBA, StateBB) {
        STATE(StateBBA)

    private:
        void entry() { TOP::box().entry(STATE_BBA); }
        void exit() { TOP::box().exit(STATE_BBA); }
    };

    SUBSTATE(StateBBB, StateBB) {
        STATE(StateBBB)

    private:
        void entry() { TOP::box().entry(STATE_BBB); }
        void exit() { TOP::box().exit(STATE_BBB); }
    };

    SUBSTATE(StateC, Top) {
        STATE(StateC)
        DEEPHISTORY()

    private:
        void init();
        void entry() { TOP::box().entry(STATE_C); }
        void exit() { TOP::box().exit(STATE_C); }
    };

    SUBSTATE(StateCA, StateC) {
        STATE(StateCA)
        DEEPHISTORY()

    private:
        void init();
        void entry() { TOP::box().entry(STATE_CA); }
        void exit() { TOP::box().exit(STATE_CA); }
    };

    SUBSTATE(StateCAA, StateCA) {
        STATE(StateCAA)

    private:
        void entry() { TOP::box().entry(STATE_CAA); }
        void exit() { TOP::box().exit(STATE_CAA); }
    };

    SUBSTATE(StateCAB, StateCA) {
        STATE(StateCAB)

    private:
        void entry() { TOP::box().entry(STATE_CAB); }
        void exit() { TOP::box().exit(STATE_CAB); }
    };

    SUBSTATE(StateCB, StateC) {
        STATE(StateCB)

    private:
        void entry() { TOP::box().entry(STATE_CB); }
        void exit() { TOP::box().exit(STATE_CB); }
    };

    SUBSTATE(StateCBA, StateCB) {
        STATE(StateCBA)

    private:
        void entry() { TOP::box().entry(STATE_CBA); }
        void exit() { TOP::box().exit(STATE_CBA); }
    };

    SUBSTATE(StateCBB, StateCB) {
        STATE(StateCBB)

    private:
        void entry() { TOP::box().entry(STATE_CBB); }
        void exit() { TOP::box().exit(STATE_CBB); }
    };

    void StateC::init() { setState<StateCA>(); }
    void StateCA::init() { setState<StateCAA>(); }
}

////////////////////////////////////////////////////////////////////////////////
// Testing dispatch mechanism
enum {
    EVENT1,
    EVENT2,
    EVENT3,
    STATEA_ENTRY,
    STATEB_ENTRY
};

namespace Dispatch {

    TOPSTATE(Top) {
        typedef std::vector<int> Box;

        STATE(Top)

        virtual void clear() {
            box().clear();
        }
        virtual void event1(int i) {
            assert(i = 1);
            box().push_back(EVENT1);
        }
        virtual void event2(int i, bool b) {
            assert(i = 2); assert(!b);
            box().push_back(EVENT2);
        }
        virtual void event3(int i, bool b) {
            assert(i = 3); assert(b);
            box().push_back(EVENT3);
        }
    };

    SUBSTATE(StateA, Top) {
        STATE(StateA)

        virtual void event3(int i, bool b);
    private:
        void entry() {
            Top::box().push_back(STATEA_ENTRY);
        }
    };

    SUBSTATE(StateB, Top) {
        STATE(StateB)

        virtual void event3(int i, bool b);
    private:
        void entry() {
            Top::box().push_back(STATEB_ENTRY);
        }
    };

    void StateA::event3(int i, bool b) {
        SUPER::event3(i, b);
        setState<StateB>();
        dispatch(Event(&Top::event1, 1));
    }

    void StateB::event3(int i, bool b) {
        SUPER::event3(i, b);
        dispatch(Event(&Top::event1, 1));
        setState<StateA>();
    }
}

////////////////////////////////////////////////////////////////////////////////
// Helper functions
template<typename T>
void setState(Macho::Machine<typename T::TOP> & m, typename T::Box * box = 0) {
    m.setState(T::_getInfo(&m), true, box);
}

template<typename T>
void setStateDirect(Macho::Machine<typename T::TOP> & m, typename T::Box * box = 0) {
    m.setState(T::_getInfo(&m), false, box);
}

template<class T>
bool checkBox(Macho::Machine<Transitions::Top> & m) {
    // Will assert if box is missing
    static_cast<T*>(m.myCurrentState->instance())->T::box();
    return true;
}

////////////////////////////////////////////////////////////////////////////////
// These tests will test for proper entry/exit, box and history handling
void testTransitions() {
    Macho::Machine<Transitions::Top> m;

    assert(m.box().entries == STATE_TOP); assert(m.box().exits == 0);
    m.box().clear();

    m.box().data = 42;

    // Parent -> Child
    setState<Transitions::StateB>(m);
    assert(Transitions::StateB::isCurrent(m));
    assert(Transitions::StateB::isCurrentDirect(m));
    assert(m.box().entries == STATE_B); assert(m.box().exits == 0);
    assert(checkBox<Transitions::StateB>(m));
    m.box().clear();

    // Self transition with History
    setState<Transitions::StateB>(m);
    assert(Transitions::StateB::isCurrent(m));
    assert(Transitions::StateB::isCurrentDirect(m));
    assert(m.box().entries == STATE_B); assert(m.box().exits == STATE_B);
    assert(checkBox<Transitions::StateB>(m));
    m.box().clear();

    // Sibling -> Sibling
    setState<Transitions::StateA>(m);
    assert(Transitions::StateA::isCurrent(m));
    assert(Transitions::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == STATE_B);
    assert(checkBox<Transitions::StateA>(m));
    m.box().clear();

    // Self transition
    setState<Transitions::StateA>(m);
    assert(Transitions::StateA::isCurrent(m));
    assert(Transitions::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == STATE_A);
    assert(checkBox<Transitions::StateA>(m));
    m.box().clear();

    // Parent -> Child
    setState<Transitions::StateAA>(m);
    assert(Transitions::StateA::isCurrent(m));
    assert(!Transitions::StateA::isCurrentDirect(m));
    assert(Transitions::StateAA::isCurrent(m));
    assert(Transitions::StateAA::isCurrentDirect(m));
    assert(m.box().entries == STATE_AA); assert(m.box().exits == 0);
    assert(checkBox<Transitions::StateA>(m));
    assert(checkBox<Transitions::StateAA>(m));
    m.box().clear();

    // Child -> Parent
    setState<Transitions::StateA>(m);
    assert(Transitions::StateA::isCurrent(m));
    assert(Transitions::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == (STATE_AA | STATE_A) );
    assert(checkBox<Transitions::StateA>(m));
    m.box().clear();

    // Parent -> Child*
    setState<Transitions::StateAAA>(m);
    assert(Transitions::StateA::isCurrent(m));
    assert(!Transitions::StateA::isCurrentDirect(m));
    assert(Transitions::StateAA::isCurrent(m));
    assert(!Transitions::StateAA::isCurrentDirect(m));
    assert(Transitions::StateAAA::isCurrent(m));
    assert(Transitions::StateAAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_AA | STATE_AAA) ); assert(m.box().exits == 0);
    assert(checkBox<Transitions::StateA>(m));
    assert(checkBox<Transitions::StateAA>(m));
    assert(checkBox<Transitions::StateAAA>(m));
    m.box().clear();

    // Sibling* -> Sibling*
    setState<Transitions::StateABB>(m);
    assert(Transitions::StateA::isCurrent(m));
    assert(!Transitions::StateA::isCurrentDirect(m));
    assert(Transitions::StateAB::isCurrent(m));
    assert(!Transitions::StateAB::isCurrentDirect(m));
    assert(Transitions::StateABB::isCurrent(m));
    assert(Transitions::StateABB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_AB | STATE_ABB) ); assert(m.box().exits == (STATE_AAA | STATE_AA) );
    assert(checkBox<Transitions::StateA>(m));
    assert(checkBox<Transitions::StateAB>(m));
    assert(checkBox<Transitions::StateABB>(m));
    m.box().clear();

    // Child -> Parent*
    setState<Transitions::StateA>(m);
    assert(Transitions::StateA::isCurrent(m));
    assert(Transitions::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == (STATE_ABB | STATE_AB | STATE_A) );
    assert(checkBox<Transitions::StateA>(m));
    m.box().clear();

    // Sibling -> Child* of Sibling
    setState<Transitions::StateBAA>(m);
    assert(Transitions::StateB::isCurrent(m));
    assert(!Transitions::StateB::isCurrentDirect(m));
    assert(Transitions::StateBA::isCurrent(m));
    assert(!Transitions::StateBA::isCurrentDirect(m));
    assert(Transitions::StateBAA::isCurrent(m));
    assert(Transitions::StateBAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_B | STATE_BA | STATE_BAA) ); assert(m.box().exits == STATE_A);
    assert(checkBox<Transitions::StateB>(m));
    assert(checkBox<Transitions::StateBA>(m));
    assert(checkBox<Transitions::StateBAA>(m));
    m.box().clear();

    // Child* of Sibling -> Sibling with Init
    setState<Transitions::StateC>(m);
    assert(Transitions::StateC::isCurrent(m));
    assert(!Transitions::StateC::isCurrentDirect(m));
    assert(Transitions::StateCA::isCurrent(m));
    assert(!Transitions::StateCA::isCurrentDirect(m));
    assert(Transitions::StateCAA::isCurrent(m));
    assert(Transitions::StateCAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAA) ); assert(m.box().exits == (STATE_BAA | STATE_BA | STATE_B) );
    assert(checkBox<Transitions::StateC>(m));
    assert(checkBox<Transitions::StateCA>(m));
    assert(checkBox<Transitions::StateCAA>(m));
    m.box().clear();

    // Child* of Sibling -> Sibling with Shallow History
    setState<Transitions::StateB>(m);
    assert(Transitions::StateB::isCurrent(m));
    assert(!Transitions::StateB::isCurrentDirect(m));
    assert(Transitions::StateBA::isCurrent(m));
    assert(Transitions::StateBA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_B | STATE_BA) ); assert(m.box().exits == (STATE_CAA | STATE_CA | STATE_C) );
    assert(checkBox<Transitions::StateB>(m));
    assert(checkBox<Transitions::StateBA>(m));
    m.box().clear();

    // Child of Sibling -> Sibling with Deep History
    setState<Transitions::StateC>(m);
    assert(Transitions::StateC::isCurrent(m));
    assert(!Transitions::StateC::isCurrentDirect(m));
    assert(Transitions::StateCA::isCurrent(m));
    assert(!Transitions::StateCA::isCurrentDirect(m));
    assert(Transitions::StateCAA::isCurrent(m));
    assert(Transitions::StateCAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAA) ); assert(m.box().exits == (STATE_BA | STATE_B) );
    assert(checkBox<Transitions::StateC>(m));
    assert(checkBox<Transitions::StateCA>(m));
    assert(checkBox<Transitions::StateCAA>(m));
    m.box().clear();

    // Child -> Parent with Deep History
    setState<Transitions::StateC>(m);
    assert(Transitions::StateC::isCurrent(m));
    assert(!Transitions::StateC::isCurrentDirect(m));
    assert(Transitions::StateCA::isCurrent(m));
    assert(!Transitions::StateCA::isCurrentDirect(m));
    assert(Transitions::StateCAA::isCurrent(m));
    assert(Transitions::StateCAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAA) ); assert(m.box().exits == (STATE_C | STATE_CA | STATE_CAA) );
    assert(checkBox<Transitions::StateC>(m));
    assert(checkBox<Transitions::StateCA>(m));
    assert(checkBox<Transitions::StateCAA>(m));
    m.box().clear();

    // Sibling -> Sibling, then Child -> Parent with Deep History
    setState<Transitions::StateCAB>(m);
    assert(Transitions::StateC::isCurrent(m));
    assert(!Transitions::StateC::isCurrentDirect(m));
    assert(Transitions::StateCA::isCurrent(m));
    assert(!Transitions::StateCA::isCurrentDirect(m));
    assert(Transitions::StateCAB::isCurrent(m));
    assert(Transitions::StateCAB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_CAB) ); assert(m.box().exits == (STATE_CAA) );
    assert(checkBox<Transitions::StateC>(m));
    assert(checkBox<Transitions::StateCA>(m));
    assert(checkBox<Transitions::StateCAB>(m));
    m.box().clear();

    setState<Transitions::StateC>(m);
    assert(Transitions::StateC::isCurrent(m));
    assert(!Transitions::StateC::isCurrentDirect(m));
    assert(Transitions::StateCA::isCurrent(m));
    assert(!Transitions::StateCA::isCurrentDirect(m));
    assert(Transitions::StateCAB::isCurrent(m));
    assert(Transitions::StateCAB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAB) ); assert(m.box().exits == (STATE_C | STATE_CA | STATE_CAB) );
    assert(checkBox<Transitions::StateC>(m));
    assert(checkBox<Transitions::StateCA>(m));
    assert(checkBox<Transitions::StateCAB>(m));
    m.box().clear();

    // Sibling* -> Sibling*
    setState<Transitions::StateCBB>(m);
    assert(Transitions::StateC::isCurrent(m));
    assert(!Transitions::StateC::isCurrentDirect(m));
    assert(Transitions::StateCB::isCurrent(m));
    assert(!Transitions::StateCB::isCurrentDirect(m));
    assert(Transitions::StateCBB::isCurrent(m));
    assert(Transitions::StateCBB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_CB | STATE_CBB) ); assert(m.box().exits == (STATE_CAB | STATE_CA) );
    assert(checkBox<Transitions::StateC>(m));
    assert(checkBox<Transitions::StateCB>(m));
    assert(checkBox<Transitions::StateCBB>(m));
    m.box().clear();

    // Child* of Sibling -> Sibling
    setState<Transitions::StateA>(m);
    assert(Transitions::StateA::isCurrent(m));
    assert(Transitions::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == (STATE_CBB | STATE_CB | STATE_C) );
    assert(checkBox<Transitions::StateA>(m));
    m.box().clear();

    // Sibling -> Sibling with Deep History
    setState<Transitions::StateC>(m);
    assert(Transitions::StateC::isCurrent(m));
    assert(!Transitions::StateC::isCurrentDirect(m));
    assert(Transitions::StateCB::isCurrent(m));
    assert(!Transitions::StateCB::isCurrentDirect(m));
    assert(Transitions::StateCBB::isCurrent(m));
    assert(Transitions::StateCBB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CB | STATE_CBB) ); assert(m.box().exits == STATE_A);
    assert(checkBox<Transitions::StateC>(m));
    assert(checkBox<Transitions::StateCB>(m));
    assert(checkBox<Transitions::StateCBB>(m));
    m.box().clear();

    // Ignoring history
    setStateDirect<Transitions::StateC>(m);
    assert(Transitions::StateC::isCurrent(m));
    assert(!Transitions::StateC::isCurrentDirect(m));
    assert(Transitions::StateCA::isCurrent(m));
    assert(!Transitions::StateCA::isCurrentDirect(m));
    assert(Transitions::StateCAB::isCurrent(m));
    assert(Transitions::StateCAB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAB) ); assert(m.box().exits == (STATE_C | STATE_CB | STATE_CBB));
    assert(checkBox<Transitions::StateC>(m));
    assert(checkBox<Transitions::StateCA>(m));
    assert(checkBox<Transitions::StateCAA>(m));
    m.box().clear();

    // Clearing history
    Transitions::StateB::clearHistory(m);
    setState<Transitions::StateB>(m);
    assert(Transitions::StateB::isCurrent(m));
    assert(Transitions::StateB::isCurrentDirect(m));
    assert(m.box().entries == STATE_B); assert(m.box().exits == (STATE_C | STATE_CA | STATE_CAB));
    assert(checkBox<Transitions::StateB>(m));

    // Clearing History deep
    setState<Transitions::StateCAB>(m);
    setState<Transitions::StateB>(m);
    m.box().clear();

    Transitions::StateC::clearHistoryDeep(m);
    setState<Transitions::StateC>(m);
    assert(Transitions::StateC::isCurrent(m));
    assert(!Transitions::StateC::isCurrentDirect(m));
    assert(Transitions::StateCA::isCurrent(m));
    assert(!Transitions::StateCA::isCurrentDirect(m));
    assert(Transitions::StateCAA::isCurrent(m));
    assert(Transitions::StateCAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAA) ); assert(m.box().exits == (STATE_B));
    assert(checkBox<Transitions::StateC>(m));
    assert(checkBox<Transitions::StateCA>(m));
    assert(checkBox<Transitions::StateCAA>(m));
    //assert(checkBox<Transitions::StateCAB>(m));
    m.box().clear();

    assert(m.box().data == 42);
}

////////////////////////////////////////////////////////////////////////////////
// Testing dispatch mechanism
void testDispatch() {
    Macho::Machine<Dispatch::Top> m;

    // Test dispatching of events
    m.dispatch(Event(&Dispatch::Top::event1, 1));
    m.dispatch(Event(&Dispatch::Top::event2, 2, false));
    m.dispatch(Event(&Dispatch::Top::event3, 3, true));

    assert(m.box()[0] == EVENT1); assert(m.box()[1] == EVENT2); assert(m.box()[2] == EVENT3);
    m->clear();

    setState<Dispatch::StateA>(m);
    m->clear();
    // Test if internal dispatch happens after switching to new state
    m.dispatch(Event(&Dispatch::Top::event3, 3, true));
    assert(m.box()[0] == EVENT3); assert(m.box()[1] == STATEB_ENTRY); assert(m.box()[2] == EVENT1);

    m->clear();
    // Test if internal dispatch happens after switching to new state
    m.dispatch(Event(&Dispatch::Top::event3, 3, true));
    assert(m.box()[0] == EVENT3); assert(m.box()[1] == STATEA_ENTRY); assert(m.box()[2] == EVENT1);
}

int main() {
    cout << "Testing transitions" << endl;
    testTransitions();

    cout << endl << "Testing dispatch mechanism" << endl;
    testDispatch();

    return 0;
}
