
#include "Macho.hpp"

#include <iostream>
using namespace std;

namespace Example {
    TOPSTATE(Top) {
        struct Box { 
            Box() : myData(0) {} 
            Box(long l) : myData(l) {} 
            long myData; 
        };

        STATE(Top)

        virtual void event1(int i) {}
        virtual void event2(long l) {}
    
    private:
        void entry();
        void exit();
        void init();
    };
    
    SUBSTATE(Super, Top) {
        STATE(Super)
        HISTORY()
    
    private:        
        void entry();
        void exit();
    };

    SUBSTATE(StateA, Super) {
        struct Box { 
            Box() : myData(0) {} 
            Box(int i) : myData(i) {} 
            int myData; 
        };
        
        STATE(StateA)

        void event1(int i);

    private:
        void entry();
        void exit();        
    };
    
    SUBSTATE(StateB, Super) {
        STATE(StateB)

        void event2(long l);

    private:
        void entry();
        void exit();
    };
    
    void Top::entry() { cout << "Top::entry" << endl; }
    void Top::exit() { cout << "Top::exit" << endl; }   
    void Top::init() {
        // Initialize state with box
        setState(StateA::key(), new StateA::Box(44));
    }

    void Super::entry() { cout << "Super::entry" << endl; }
    void Super::exit() { cout << "Super::exit" << endl; }   

    void StateA::entry() { cout << "StateA::entry" << endl; }
    void StateA::exit() { cout << "StateA::exit" << endl; }
    void StateA::event1(int i) {
        box().myData = i;
        cout << "StateA::box().myData: " << box().myData << endl;
        setState(StateB::key());
    }

    void StateB::entry() { cout << "StateB::entry" << endl; }
    void StateB::exit() { cout << "StateB::exit" << endl; }         
    void StateB::event2(long l) {
        Top::box().myData = l;
        cout << "Top::box().myData: " << Top::box().myData << endl;
        setState(StateA::key());
    }
}

int main() {
    // Initialize state machine with box
    Macho::Machine<Example::Top> m(new Example::Top::Box(11));
    m->event1(42);
    m->event2(43);

    cout << "m.box().myData: " << m.box().myData << endl;
    
    return 0;

/*
   Output is:

   Top::entry
   Super::entry
   StateA::entry
   StateA::box().myData: 42
   StateA::exit
   StateB::entry
   Top::box().myData: 43
   StateB::exit
   StateA::entry
   m.box().myData: 43
   StateA::exit
   Super::exit
   Top::exit
*/
}
