
// need to open up Machine class to make tests run
#define protected public
#define private public

// lets see everything that happens
#define __MACHO_TRACE__

#include "Macho.hpp"

enum {
    STATE_TOP = 1 << 0,
    STATE_A = 1 << 1,
    STATE_AA = 1 << 2,
    STATE_AB = 1 << 3,
    STATE_AAA = 1 << 4,
    STATE_AAB = 1 << 5,
    STATE_ABA = 1 << 6,
    STATE_ABB = 1 << 7,
    STATE_B = 1 << 8,
    STATE_BA = 1 << 9,
    STATE_BB = 1 << 10,
    STATE_BAA = 1 << 11,
    STATE_BAB = 1 << 12,
    STATE_BBA = 1 << 13,
    STATE_BBB = 1 << 14,
    STATE_C = 1 << 15,
    STATE_CA = 1 << 16,
    STATE_CB = 1 << 17,
    STATE_CAA = 1 << 18,
    STATE_CAB = 1 << 19,
    STATE_CBA = 1 << 20,
    STATE_CBB = 1 << 21
};


// These tests will test for proper entry/exit, box and history handling
namespace Test {

    TOPSTATE(Top) {
        struct Box {
            Box() : data(0) { clear(); }

            // Registering all exits and entries
            void entry(long state) { entries |= state; }
            void exit(long state) { exits |= state; }
            void clear() const { entries = 0; exits = 0; }

            mutable long entries;
            mutable long exits;
            mutable int data;
        };

        STATE(Top)

    private:
        void entry() { box().entry(STATE_TOP); }
        void exit() { box().exit(STATE_TOP); }
    };

    SUBSTATE(StateA, Top) {
        struct Box {
            Box() : data(0) {}
            int data;
        };

        STATE(StateA)

    private:
        // Testing proper box handling here
        void entry() { TOP::box().entry(STATE_A); box().data = 42; }
        void exit() { TOP::box().exit(STATE_A); assert(box().data == 42); }
    };

    SUBSTATE(StateAA, StateA) {
        STATE(StateAA)

    private:
        // Testing proper box handling here
        void entry() { TOP::box().entry(STATE_AA); assert(StateA::box().data == 42); StateA::box().data = 43; }
        void exit() { TOP::box().exit(STATE_AA); assert(StateA::box().data == 43); StateA::box().data = 42; }
    };

    SUBSTATE(StateAAA, StateAA) {
        STATE(StateAAA)

    private:
        void entry() { TOP::box().entry(STATE_AAA); }
        void exit() { TOP::box().exit(STATE_AAA); }
    };

    SUBSTATE(StateAAB, StateAA) {
        STATE(StateAAB)

    private:
        void entry() { TOP::box().entry(STATE_AAB); }
        void exit() { TOP::box().exit(STATE_AAB); }
    };

    SUBSTATE(StateAB, StateA) {
        STATE(StateAB)

    private:
        void entry() { TOP::box().entry(STATE_AB); }
        void exit() { TOP::box().exit(STATE_AB); }
    };

    SUBSTATE(StateABA, StateAB) {
        STATE(StateABA)

    private:
        void entry() { TOP::box().entry(STATE_ABA); }
        void exit() { TOP::box().exit(STATE_ABA); }
    };

    SUBSTATE(StateABB, StateAB) {
        STATE(StateABB)

    private:
        void entry() { TOP::box().entry(STATE_ABB); }
        void exit() { TOP::box().exit(STATE_ABB); }
    };

    SUBSTATE(StateB, Top) {
        STATE(StateB)
        HISTORY()

    private:
        void entry() { TOP::box().entry(STATE_B); }
        void exit() { TOP::box().exit(STATE_B); }
    };

    SUBSTATE(StateBA, StateB) {
        STATE(StateBA)

    private:
        void entry() { TOP::box().entry(STATE_BA); }
        void exit() { TOP::box().exit(STATE_BA); }
    };

    SUBSTATE(StateBAA, StateBA) {
        STATE(StateBAA)

    private:
        void entry() { TOP::box().entry(STATE_BAA); }
        void exit() { TOP::box().exit(STATE_BAA); }
    };

    SUBSTATE(StateBAB, StateBA) {
        STATE(StateBAB)

    private:
        void entry() { TOP::box().entry(STATE_BAB); }
        void exit() { TOP::box().exit(STATE_BAB); }
    };

    SUBSTATE(StateBB, StateB) {
        STATE(StateBB)

    private:
        void entry() { TOP::box().entry(STATE_BB); }
        void exit() { TOP::box().exit(STATE_BB); }
    };

    SUBSTATE(StateBBA, StateBB) {
        STATE(StateBBA)

    private:
        void entry() { TOP::box().entry(STATE_BBA); }
        void exit() { TOP::box().exit(STATE_BBA); }
    };

    SUBSTATE(StateBBB, StateBB) {
        STATE(StateBBB)

    private:
        void entry() { TOP::box().entry(STATE_BBB); }
        void exit() { TOP::box().exit(STATE_BBB); }
    };

    SUBSTATE(StateC, Top) {
        STATE(StateC)
        DEEPHISTORY()

    private:
        void init();
        void entry() { TOP::box().entry(STATE_C); }
        void exit() { TOP::box().exit(STATE_C); }
    };

    SUBSTATE(StateCA, StateC) {
        STATE(StateCA)
        DEEPHISTORY()

    private:
        void init();
        void entry() { TOP::box().entry(STATE_CA); }
        void exit() { TOP::box().exit(STATE_CA); }
    };

    SUBSTATE(StateCAA, StateCA) {
        STATE(StateCAA)

    private:
        void entry() { TOP::box().entry(STATE_CAA); }
        void exit() { TOP::box().exit(STATE_CAA); }
    };

    SUBSTATE(StateCAB, StateCA) {
        STATE(StateCAB)

    private:
        void entry() { TOP::box().entry(STATE_CAB); }
        void exit() { TOP::box().exit(STATE_CAB); }
    };

    SUBSTATE(StateCB, StateC) {
        STATE(StateCB)

    private:
        void entry() { TOP::box().entry(STATE_CB); }
        void exit() { TOP::box().exit(STATE_CB); }
    };

    SUBSTATE(StateCBA, StateCB) {
        STATE(StateCBA)

    private:
        void entry() { TOP::box().entry(STATE_CBA); }
        void exit() { TOP::box().exit(STATE_CBA); }
    };

    SUBSTATE(StateCBB, StateCB) {
        STATE(StateCBB)

    private:
        void entry() { TOP::box().entry(STATE_CBB); }
        void exit() { TOP::box().exit(STATE_CBB); }
    };

    void StateC::init() { setState(StateCA::key()); }
    void StateCA::init() { setState(StateCAA::key()); }
}

int main() {
    Macho::Machine<Test::Top> m;

    assert(m.box().entries == STATE_TOP); assert(m.box().exits == 0);
    m.box().clear();

    m.box().data = 42;

    // Parent -> Child
    m.setStateTest(Test::StateB::key());
    assert(Test::StateB::isCurrent(m));
    assert(Test::StateB::isCurrentDirect(m));
    assert(m.box().entries == STATE_B); assert(m.box().exits == 0);
    m.box().clear();

    // Self transition with History
    m.setStateTest(Test::StateB::key());
    assert(Test::StateB::isCurrent(m));
    assert(Test::StateB::isCurrentDirect(m));
    assert(m.box().entries == STATE_B); assert(m.box().exits == STATE_B);
    m.box().clear();

    // Sibling -> Sibling
    m.setStateTest(Test::StateA::key());
    assert(Test::StateA::isCurrent(m));
    assert(Test::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == STATE_B);
    m.box().clear();

    // Self transition
    m.setStateTest(Test::StateA::key());
    assert(Test::StateA::isCurrent(m));
    assert(Test::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == STATE_A);
    m.box().clear();

    // Parent -> Child
    m.setStateTest(Test::StateAA::key());
    assert(Test::StateA::isCurrent(m));
    assert(!Test::StateA::isCurrentDirect(m));
    assert(Test::StateAA::isCurrent(m));
    assert(Test::StateAA::isCurrentDirect(m));
    assert(m.box().entries == STATE_AA); assert(m.box().exits == 0);
    m.box().clear();

    // Child -> Parent
    m.setStateTest(Test::StateA::key());
    assert(Test::StateA::isCurrent(m));
    assert(Test::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == (STATE_AA | STATE_A) );
    m.box().clear();

    // Parent -> Child*
    m.setStateTest(Test::StateAAA::key());
    assert(Test::StateA::isCurrent(m));
    assert(!Test::StateA::isCurrentDirect(m));
    assert(Test::StateAA::isCurrent(m));
    assert(!Test::StateAA::isCurrentDirect(m));
    assert(Test::StateAAA::isCurrent(m));
    assert(Test::StateAAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_AA | STATE_AAA) ); assert(m.box().exits == 0);
    m.box().clear();

    // Sibling* -> Sibling*
    m.setStateTest(Test::StateABB::key());
    assert(Test::StateA::isCurrent(m));
    assert(!Test::StateA::isCurrentDirect(m));
    assert(Test::StateAB::isCurrent(m));
    assert(!Test::StateAB::isCurrentDirect(m));
    assert(Test::StateABB::isCurrent(m));
    assert(Test::StateABB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_AB | STATE_ABB) ); assert(m.box().exits == (STATE_AAA | STATE_AA) );
    m.box().clear();

    // Child -> Parent*
    m.setStateTest(Test::StateA::key());
    assert(Test::StateA::isCurrent(m));
    assert(Test::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == (STATE_ABB | STATE_AB | STATE_A) );
    m.box().clear();

    // Sibling -> Child* of Sibling
    m.setStateTest(Test::StateBAA::key());
    assert(Test::StateB::isCurrent(m));
    assert(!Test::StateB::isCurrentDirect(m));
    assert(Test::StateBA::isCurrent(m));
    assert(!Test::StateBA::isCurrentDirect(m));
    assert(Test::StateBAA::isCurrent(m));
    assert(Test::StateBAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_B | STATE_BA | STATE_BAA) ); assert(m.box().exits == STATE_A);
    m.box().clear();

    // Child* of Sibling -> Sibling with Init
    m.setStateTest(Test::StateC::key());
    assert(Test::StateC::isCurrent(m));
    assert(!Test::StateC::isCurrentDirect(m));
    assert(Test::StateCA::isCurrent(m));
    assert(!Test::StateCA::isCurrentDirect(m));
    assert(Test::StateCAA::isCurrent(m));
    assert(Test::StateCAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAA) ); assert(m.box().exits == (STATE_BAA | STATE_BA | STATE_B) );
    m.box().clear();

    // Child* of Sibling -> Sibling with Shallow History
    m.setStateTest(Test::StateB::key());
    assert(Test::StateB::isCurrent(m));
    assert(!Test::StateB::isCurrentDirect(m));
    assert(Test::StateBA::isCurrent(m));
    assert(Test::StateBA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_B | STATE_BA) ); assert(m.box().exits == (STATE_CAA | STATE_CA | STATE_C) );
    m.box().clear();

    // Child of Sibling -> Sibling with Deep History
    m.setStateTest(Test::StateC::key());
    assert(Test::StateC::isCurrent(m));
    assert(!Test::StateC::isCurrentDirect(m));
    assert(Test::StateCA::isCurrent(m));
    assert(!Test::StateCA::isCurrentDirect(m));
    assert(Test::StateCAA::isCurrent(m));
    assert(Test::StateCAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAA) ); assert(m.box().exits == (STATE_BA | STATE_B) );
    m.box().clear();

    // Child -> Parent with Deep History
    m.setStateTest(Test::StateC::key());
    assert(Test::StateC::isCurrent(m));
    assert(!Test::StateC::isCurrentDirect(m));
    assert(Test::StateCA::isCurrent(m));
    assert(!Test::StateCA::isCurrentDirect(m));
    assert(Test::StateCAA::isCurrent(m));
    assert(Test::StateCAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAA) ); assert(m.box().exits == (STATE_C | STATE_CA | STATE_CAA) );
    m.box().clear();

    // Sibling -> Sibling, then Child -> Parent with Deep History
    m.setStateTest(Test::StateCAB::key());
    assert(Test::StateC::isCurrent(m));
    assert(!Test::StateC::isCurrentDirect(m));
    assert(Test::StateCA::isCurrent(m));
    assert(!Test::StateCA::isCurrentDirect(m));
    assert(Test::StateCAB::isCurrent(m));
    assert(Test::StateCAB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_CAB) ); assert(m.box().exits == (STATE_CAA) );
    m.box().clear();

    m.setStateTest(Test::StateC::key());
    assert(Test::StateC::isCurrent(m));
    assert(!Test::StateC::isCurrentDirect(m));
    assert(Test::StateCA::isCurrent(m));
    assert(!Test::StateCA::isCurrentDirect(m));
    assert(Test::StateCAB::isCurrent(m));
    assert(Test::StateCAB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAB) ); assert(m.box().exits == (STATE_C | STATE_CA | STATE_CAB) );
    m.box().clear();

    // Sibling* -> Sibling*
    m.setStateTest(Test::StateCBB::key());
    assert(Test::StateC::isCurrent(m));
    assert(!Test::StateC::isCurrentDirect(m));
    assert(Test::StateCB::isCurrent(m));
    assert(!Test::StateCB::isCurrentDirect(m));
    assert(Test::StateCBB::isCurrent(m));
    assert(Test::StateCBB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_CB | STATE_CBB) ); assert(m.box().exits == (STATE_CAB | STATE_CA) );
    m.box().clear();

    // Child* of Sibling -> Sibling
    m.setStateTest(Test::StateA::key());
    assert(Test::StateA::isCurrent(m));
    assert(Test::StateA::isCurrentDirect(m));
    assert(m.box().entries == STATE_A); assert(m.box().exits == (STATE_CBB | STATE_CB | STATE_C) );
    m.box().clear();

    // Sibling -> Sibling with Deep History
    m.setStateTest(Test::StateC::key());
    assert(Test::StateC::isCurrent(m));
    assert(!Test::StateC::isCurrentDirect(m));
    assert(Test::StateCB::isCurrent(m));
    assert(!Test::StateCB::isCurrentDirect(m));
    assert(Test::StateCBB::isCurrent(m));
    assert(Test::StateCBB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CB | STATE_CBB) ); assert(m.box().exits == STATE_A);
    m.box().clear();

    // Ignoring history
    m.setStateTest(Test::StateC::key(), false);
    assert(Test::StateC::isCurrent(m));
    assert(!Test::StateC::isCurrentDirect(m));
    assert(Test::StateCA::isCurrent(m));
    assert(!Test::StateCA::isCurrentDirect(m));
    assert(Test::StateCAB::isCurrent(m));
    assert(Test::StateCAB::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAB) ); assert(m.box().exits == (STATE_C | STATE_CB | STATE_CBB));
    m.box().clear();

    // Clearing history
    Test::StateB::clearHistory(m);
    m.setStateTest(Test::StateB::key());
    assert(Test::StateB::isCurrent(m));
    assert(Test::StateB::isCurrentDirect(m));
    assert(m.box().entries == STATE_B); assert(m.box().exits == (STATE_C | STATE_CA | STATE_CAB));

    // Clearing History deep
    m.setStateTest(Test::StateCAB::key());
    m.setStateTest(Test::StateB::key());
    m.box().clear();

    Test::StateC::clearHistoryDeep(m);
    m.setStateTest(Test::StateC::key());
    assert(Test::StateC::isCurrent(m));
    assert(!Test::StateC::isCurrentDirect(m));
    assert(Test::StateCA::isCurrent(m));
    assert(!Test::StateCA::isCurrentDirect(m));
    assert(Test::StateCAA::isCurrent(m));
    assert(Test::StateCAA::isCurrentDirect(m));
    assert(m.box().entries == (STATE_C | STATE_CA | STATE_CAA) ); assert(m.box().exits == (STATE_B));
    m.box().clear();

    assert(m.box().data == 42);

    return 0;
}
